/***************************************************************************
 *   Copyright (C) 2005 - 2007 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include <QtGui>

#include "mp3transcode.h"
#include "containerutils.h"
#include "logger.h"

#define MP3_MIN_BUFFER_SIZE 4096

struct mpstr_tag mpeg;

MP3Transcode::MP3Transcode() :
        m_decodedBufferCapacity( 32 * 1024 ),
        m_mpegInitialised( false )
{
    LOGL( 3, "Initialising MP3 Transcoding" );

    if ( !InitMP3( &mpeg ) )
    {
        LOGL( 1, "Something went wrong when initiliasing mpglib. "
            " God knows what. Chris can read the \"best documentation "
            " there is\" to find out. :P" );
    }

    // Maybe give these a good start value?  They'll grow as needed anyway.
    m_decBufSize = 1; // On my machine this grows to about 120K
    m_encBufSize = 1; // And this grows to about 6K or so.
    m_encBufReadIndex = m_encBufWriteIndex = 0;
    m_decBufReadIndex = m_decBufWriteIndex = 0;
    m_decodedBuffer = (char*)malloc( m_decBufSize );
    m_encodedBuffer = (char*)malloc( m_encBufSize );
}

MP3Transcode::~MP3Transcode()
{
    if ( m_decodedBuffer )
    {
        free( m_decodedBuffer);
        m_decodedBuffer = NULL;
    }
    if ( m_encodedBuffer )
    {
        free( m_encodedBuffer);
        m_encodedBuffer = NULL;
    }
    ExitMP3( &mpeg );
}


QStringList
MP3Transcode::supportedMimeTypes() const
{
    return QStringList( "application/x-mp3" );
}

QStringList
MP3Transcode::supportedFileExtensions() const
{
    return QStringList( "mp3" );
}

void
MP3Transcode::expandRingBuffer( char** buffer, int& bufSize, int& readIndex, int& writeIndex, int growSize )
{
    //qDebug() << "resizing buffer from" << bufSize
    //         << "to" << bufSize + growSize;
    char* tmp;
    tmp = (char*)realloc( *buffer, sizeof(char) * ( bufSize + growSize ) );
    if ( tmp == NULL )
    {
        Q_ASSERT( !"Could not reallocate buffer!" );
        return;
    }
    else
        *buffer = tmp;

    // Don't screw up the read order or ugly noises will ensue
    if ( readIndex > writeIndex )
    {
        memmove( *buffer + readIndex + growSize,
                 *buffer + readIndex,
                 bufSize - readIndex );
        readIndex += growSize;
    }
    bufSize += growSize;
}


bool
MP3Transcode::processData( const QByteArray &buffer )
{
    int length = buffer.size();
    int cnt;
    const char* src = buffer.data();
    int neededSize = getEncodedBufferFilled() + buffer.size();

    if ( neededSize >= m_encBufSize )
        expandRingBuffer( &m_encodedBuffer, m_encBufSize,
                          m_encBufReadIndex, m_encBufWriteIndex,
                          neededSize - m_encBufSize + 1 );

    while ( length > 0 )
    {
        int wr;
        cnt = qMin( length, m_encBufSize - m_encBufWriteIndex );
        memcpy( m_encodedBuffer + m_encBufWriteIndex, src, cnt );
        wr = (m_encBufWriteIndex + cnt) % m_encBufSize;
        m_encBufWriteIndex = wr;
        length -= cnt;
        src += cnt;
    }

    // wait till we got a minimal buffer of data, which we need for mpglib
    // being able to properly read the audio metadata (samplerate, channels...)
    if ( getEncodedBufferFilled() <= MP3_MIN_BUFFER_SIZE )
    {
        return true;
    }

    char tempBuffer[16384];
    int size;
    unsigned char inTemp[MP3_MIN_BUFFER_SIZE];
    length = MP3_MIN_BUFFER_SIZE;
    int outLen = 0;
    while ( length > 0 )
    {
        int rd;
        cnt = qMin(length, m_encBufSize - m_encBufReadIndex);
        memcpy( inTemp + outLen, m_encodedBuffer + m_encBufReadIndex, cnt);
        rd = (m_encBufReadIndex + cnt) % m_encBufSize;
        m_encBufReadIndex = rd;
        length -= cnt;
        outLen += cnt;
    }


    int result = decodeMP3(
        &mpeg, // handle
        inTemp, // in-data
        MP3_MIN_BUFFER_SIZE, // size of in-data
        tempBuffer, // out-data
        sizeof( tempBuffer ), // size of out-data
        &size ); // done size

    if ( result == MP3_ERR )
    {
        LOG( 1, "decodeMP3 failed. result: " << result );
        return false;
    }

    if ( !m_mpegInitialised )
    {
        long sampleRate = freqs[mpeg.fr.sampling_frequency];
        int channels = mpeg.fr.stereo > 0 ? mpeg.fr.stereo : 2;

        LOGL( 4, "mpegTranscode (Samplerate:" << sampleRate <<
                 "Channels:" << channels << ")" );

        // For certain corrupt previews, we get a result of OK, but the sample
        // rate comes out all wrong. If we let these proceed, things will crash
        // horribly.
        if ( sampleRate != 44100 || channels != 2 )
        {
            LOGL( 1, "Stream is not 44.1k stereo, aborting" );
            return false;
        }

        m_mpegInitialised = true;
        emit streamInitialized( sampleRate, channels );
    }

    while ( result == MP3_OK )
    {
        int length = size;
        int cnt;
        const char* src = tempBuffer;
        int neededSize = getDecodedBufferFilled() + size;

        if ( neededSize >= m_decBufSize )
            expandRingBuffer( &m_decodedBuffer, m_decBufSize,
                              m_decBufReadIndex, m_decBufWriteIndex,
                              ( neededSize - m_decBufSize + 1 ) );

        while (length > 0)
        {
            int wr;
            cnt = qMin(length, m_decBufSize - m_decBufWriteIndex);
            memcpy( m_decodedBuffer + m_decBufWriteIndex, src, cnt);
            wr = (m_decBufWriteIndex + cnt) % m_decBufSize;
            m_decBufWriteIndex = wr;
            length -= cnt;
            src += cnt;
        }

        result = decodeMP3( &mpeg, NULL, 0, tempBuffer, sizeof( tempBuffer ), &size );

        if ( result == MP3_ERR )
        {
            LOGL( 1, "decodeMP3 failed. result: " << result );
            return false;
        }
    }
    
    return true;
}


void
MP3Transcode::clearBuffers()
{
    ExitMP3( &mpeg );

    m_decBufReadIndex = m_decBufWriteIndex = 0;
    m_encBufReadIndex = m_encBufWriteIndex = 0;
    m_mpegInitialised = false;

    if ( !InitMP3( &mpeg ) )
    {
        LOGL( 1, "Something went wrong when initiliasing mpglib. "
            " God knows what. Chris can read the \"best documentation "
            " there is\" to find out. :P" );
    }
}

int
MP3Transcode::getEncodedBufferFilled()
{
    if ( m_encBufWriteIndex >= m_encBufReadIndex )
    {
        return m_encBufWriteIndex - m_encBufReadIndex;
    }
    return ( m_encBufSize - ( m_encBufReadIndex - m_encBufWriteIndex ) );
}

int
MP3Transcode::getDecodedBufferFilled()
{
    if ( m_decBufWriteIndex >= m_decBufReadIndex )
    {
        return m_decBufWriteIndex - m_decBufReadIndex;
    }
    return ( m_decBufSize - ( m_decBufReadIndex - m_decBufWriteIndex ) );
}

bool
MP3Transcode::needsData()
{
    return getDecodedBufferFilled() < m_decodedBufferCapacity;
}


bool
MP3Transcode::hasData()
{
    return ( getDecodedBufferFilled() > 0 );
}


void
MP3Transcode::data( QByteArray& fillMe, int numBytes )
{
    int length;
    int cnt;
    length = qMin( numBytes, ssize_t( getDecodedBufferFilled() ) );

    while (length > 0)
    {
        int rd;
        cnt = qMin (length, m_decBufSize - m_decBufReadIndex );
        fillMe.append( QByteArray::fromRawData( m_decodedBuffer + m_decBufReadIndex, cnt ) );
        rd = ( m_decBufReadIndex + cnt ) % m_decBufSize;
        m_decBufReadIndex = rd;
        length -= cnt;
    }
}


Q_EXPORT_PLUGIN2( transcode, MP3Transcode )
